/*
 * RGB-Matrix-Uhr 64x32, Scott-Falk Hhn, "functions.h"
 * Verschiedene Funktionen
 */ 

#ifndef FUNCTIONS_H_
#define FUNCTIONS_H_

void matrix_swap(void)                                                // LED-Matrix-Bank umschalten
{
  matnum = 1 - matnum;                                                // Matrix-Nummer invertieren
}

void matrix_off(void)                                                 // LED-Matrix ausschalten (fr sicheren DCF77-Empfang)
{
  if (TIMSK1) {                                                       // wenn Timer1-Interrupt eingeschaltet
    TIMSK1 = 0;                                                       // Interrupt ausschalten
    PORTA |= _BV(PINA5);                                              // OE auf High (LED-Ausgnge deaktivieren)
  }
}

void matrix_on(void)                                                  // LED-Matrix einschalten
{
  if (TIMSK1 == 0) TIMSK1 = _BV(OCIE1A);                              // wenn Timer1 Interrupt ausgeschaltet -> Timer1 Output Compare A Match Interrupt aktivieren
}

uint8_t matrix_state(void)                                            // Status der LED-Matrix ermitteln
{
  return TIMSK1;                                                      // Status = 0, wenn LED-Matrix ausgeschaltet; Status > 0, wenn LED-Matrix eingeschaltet
}

void matrix_clear(void)                                               // LED-Matrix lschen (inaktiven Teil)
{
  uint16_t i;                                                         // Bytezhler

  for (i = 0; i < sizeof(matrix[1 - matnum]); i ++) matrix[1 - matnum][i] = 0;  // Byte auf 0 setzen
}

void matrix_color(uint8_t c)                                          // Farbe fr auszugebende Pixel setzen
{                                                                     // c: Farbnummer 0-16
  color1 = pgm_read_byte(&colors[c * 2]);                             // Farbwert 1 fr Matrix aus Tabelle holen
  color2 = pgm_read_byte(&colors[c * 2 + 1]);                         // Farbwert 2 fr Matrix aus Tabelle holen
}

void matrix_pixel(int8_t x, int8_t y)                                 // Ein Pixel in der LED-Matrix mit dem aktuellen Farbwert setzen
{                                                                     // x: X-Position (0-63), y: Y-Position (0-31)
  uint16_t bufpos;                                                    // Zwischenspeicher fr berechnete Pufferposition

  if ((x >= 0) && (x < 64) && (y >= 0) && (y < 32)) {                 // wenn Positionsangaben im sichtbaren Bereich
    bufpos = (y % 16) * 64 + x;                                       // Pufferposition berechnen
    if (y < 16) {                                                     // wenn obere Bildhlfte
      matrix[1 - matnum][bufpos] |= color1 * 4;                       // Farbteil 1 setzen (inaktive Matrix)
      matrix[1 - matnum][bufpos + 1024] |= color2 * 4;                // Farbteil 2 setzen (inaktive Matrix)
    }
    else {                                                            // sonst untere Bildhlfte
      matrix[1 - matnum][bufpos] |= color1 * 32;                      // Farbteil 1 setzen (inaktive Matrix)
      matrix[1 - matnum][bufpos + 1024] |= color2 * 32;               // Farbteil 2 setzen (inaktive Matrix)
    }
  }
}

void matrix_write_reg(uint16_t val, uint8_t pos)                      // Matrix-Treiber FM6126A Register setzen (nach https://github.com/2dom/PxMatrix)
{                                                                     // val: Registerwert (16bit), pos: Registerposition
  uint8_t i;                                                          // Zhlervariable
  uint8_t e = 64;                                                     // Zhlerendwert auf 64 Bits setzen
  
  if (matscn) e = 128;                                                // wenn Matrix-Scan-Modus 1:8 -> Zhlerendwert auf 128 Bits setzen
  for (i = 0; i < e; i ++) {                                          // 64 oder 128 Bits ausgeben
    if ((val >> (15 - (i % 16))) & 1) PORTC = 0xfc;                   // wenn Bit in Registerwert gesetzt -> 1 auf allen Farbkanlen ausgeben (wiederholt sich alle 16 Bit)
    else PORTC = 0;                                                   // sonst 0 ausgeben
    PORTB = PORTB | _BV(PINB0);                                       // CLK auf High setzen
    PORTB = PORTB & ~_BV(PINB0);                                      // CLK auf Low setzen
    if (i == (e - pos)) PORTA = PORTA | _BV(PINA4);                   // wenn Zhlerstand fr Register erreicht -> LAT auf High setzen
  }
  PORTA = PORTA & ~_BV(PINA4);                                        // LAT auf Low setzen
}

void matrix_fm6126a_init(void)                                        // Matrix-Treiber FM6126A initialisieren (nach https://github.com/2dom/PxMatrix)
{
  matrix_write_reg(0b0111111111111111, 12);                           // Register 12 setzen
  matrix_write_reg(0b0000000001000000, 13);                           // Register 13 setzen
}

void color_clock(void)                                                // Uhrzeit-Farbe einstellen
{
  if (colclk == 0) matrix_color(colclkr);                             // wenn Zufallsfarbe ausgewhlt -> einstellen
  else matrix_color(colclk - 1);                                      // sonst gespeicherte Farbe einstellen
}

void color_date(void)                                                 // Datum-Farbe einstellen
{
  if (coldat == 0) matrix_color(coldatr);                             // wenn Zufallsfarbe ausgewhlt -> einstellen
  else matrix_color(coldat - 1);                                      // sonst gespeicherte Farbe einstellen
}

void color_alarm(void)                                                // Alarm-Farbe einstellen
{
  if (colalm == 0) matrix_color(colalmr);                             // wenn Zufallsfarbe ausgewhlt -> einstellen
  else matrix_color(colalm - 1);                                      // sonst gespeicherte Farbe einstellen
}

void color_random(void)                                               // Neue Zufallsfarben erzeugen
{
  uint8_t c1, c2, c3, c4;                                             // Zwischenspeicher fr bisherige Farben und zu meidende Farbe

  if (colclk == 0) c1 = colclkr;                                      // wenn Zeit-Farbe auf Zufall gesetzt -> Zufallsfarbe speichern
  else c1 = colclk - 1;                                               // sonst festen Farbwert speichern
  if (coldat == 0) c2 = coldatr;                                      // wenn Datum-Farbe auf Zufall gesetzt -> Zufallsfarbe speichern
  else c2 = coldat - 1;                                               // sonst festen Farbwert speichern
  if (colalm == 0) c3 = colalmr;                                      // wenn Alarm-Farbe auf Zufall gesetzt -> Zufallsfarbe speichern
  else c3 = colalm - 1;                                               // sonst festen Farbwert speichern
  if (avocol) c4 = avocol - 1;                                        // wenn zu meidende Farbe gesetzt -> Farbe speichern
  else c4 = 16;                                                       // sonst unglte Farbe setzen

  if (colclk == 0) {                                                  // wenn Zeit-Farbe auf Zufall gesetzt ist
    do {                                                              // Schleife
      colclkr = rand() / (RAND_MAX / 16);                             // neue Zufallsfarbe fr Uhrzeit ermitteln
    } while ((colclkr == c1) || (colclkr == c2) ||
      (colclkr == c3) || (colclkr == c4));                            // solange wiederholen, bis alle Farben unterschiedlich sind
    c1 = colclkr;                                                     // neue Zufallsfarbe speichern
  }
  if (coldat == 0) {                                                  // wenn Datum-Farbe auf Zufall gesetzt ist
    do {                                                              // Schleife
      coldatr = rand() / (RAND_MAX / 16);                             // neue Zufallsfarbe fr Datum ermitteln
    } while ((coldatr == c1) || (coldatr == c2) ||
      (coldatr == c3) || (coldatr == c4));                            // solange wiederholen, bis alle Farben unterschiedlich sind
    c2 = coldatr;                                                     // neue Zufallsfarbe speichern
  }
  if (colalm == 0) {                                                  // wenn Alarm-Farbe auf Zufall gesetzt ist
    do {                                                              // Schleife
      colalmr = rand() / (RAND_MAX / 16);                             // neue Zufallsfarbe fr Alarm ermitteln
    } while ((colalmr == c1) || (colalmr == c2) ||
      (colalmr == c3) || (colalmr == c4));                            // solange wiederholen, bis alle Farben unterschiedlich sind
  }
}

uint8_t eep_read_byte(uint16_t addr)                                  // Ein Byte von einer bestimmten EEPROM-Adresse lesen (aus AVR-GCC-Tutorial von http://www.mikrocontroller.net)
{                                                                     // addr: EEPROM-Adresse
  return eeprom_read_byte((uint8_t *) addr);                          // Byte ber die interne EEPROM-Routine lesen
}

void eep_write_byte(uint16_t addr, uint8_t val)                       // Ein Byte auf eine bestimmte EEPROM-Adresse schreiben (aus AVR-GCC-Tutorial von http://www.mikrocontroller.net)
{                                                                     // addr: EEPROM-Adresse, val: zu schreibendes Byte
  eeprom_write_byte((uint8_t *) addr, val);                           // Byte ber die interne EEPROM-Routine schreiben
}

void led_digit1(uint8_t d)                                            // Eine Groziffer 10x18 (Stunden & Minuten) in die LED-Matrix schreiben
{                                                                     // d: auszugebende Ziffer (0-9)
  uint16_t offset;                                                    // Offset in der Zeichensatz-Tabelle
  uint8_t x;                                                          // Zhler fr die Zeichenausgabe, X-Position
  uint8_t y;                                                          // Zhler fr die Zeichenausgabe, Y-Position
  uint8_t b;                                                          // Zwischenspeicher fr Bytes aus der Zeichensatz-Tabelle

  offset = d * 36;                                                    // Tabellen-Offset berechnen
  for (y = 0; y < 18; y ++) {                                         // 18 Zeilen mit je 2 Bytes ausgeben
    b = pgm_read_byte(&digit_1_array[offset ++]);                     // Zeilenbyte 1 holen
    for (x = 2; x > 0; x--) {                                         // Zeilenbyte 1 bitweise ausgeben
      if (b & 1) matrix_pixel(posx + x - 1, posy + y);                // wenn Pixel gesetzt werden muss -> LSB als Pixel ausgeben
      b >>= 1;                                                        // nchste Bit-Position an LSB
    }
    b = pgm_read_byte(&digit_1_array[offset ++]);                     // Zeilenbyte 2 holen
    for (x = 8; x > 0; x --) {                                        // Zeilenbyte 2 bitweise ausgeben
      if (b & 1) matrix_pixel(posx + x + 1, posy + y);                // wenn Pixel gesetzt werden muss -> LSB als Pixel ausgeben
      b >>= 1;                                                        // nchste Bit-Position an LSB
    }
  }
  posx += 12;                                                         // X-Position fr nchstes Zeichen
}

void led_digit2(uint8_t d)                                            // Eine Groziffer 10x8 (Stunden & Minuten) in die LED-Matrix schreiben
{                                                                     // d: auszugebende Ziffer (0-9)
  uint16_t offset;                                                    // Offset in der Zeichensatz-Tabelle
  uint8_t x;                                                          // Zhler fr die Zeichenausgabe, X-Position
  uint8_t y;                                                          // Zhler fr die Zeichenausgabe, Y-Position
  uint8_t b;                                                          // Zwischenspeicher fr Bytes aus der Zeichensatz-Tabelle

  offset = d * 16;                                                    // Tabellen-Offset berechnen
  for (y = 0; y < 8; y ++) {                                          // 8 Zeilen mit je 2 Bytes ausgeben
    b = pgm_read_byte(&digit_2_array[offset ++]);                     // Zeilenbyte 1 holen
    for (x = 2; x > 0; x--) {                                         // Zeilenbyte 1 bitweise ausgeben
      if (b & 1) matrix_pixel(posx + x - 1, posy + y);                // wenn Pixel gesetzt werden muss -> LSB als Pixel ausgeben
      b >>= 1;                                                        // nchste Bit-Position an LSB
    }
    b = pgm_read_byte(&digit_2_array[offset ++]);                     // Zeilenbyte 2 holen
    for (x = 8; x > 0; x --) {                                        // Zeilenbyte 2 bitweise ausgeben
      if (b & 1) matrix_pixel(posx + x + 1, posy + y);                // wenn Pixel gesetzt werden muss -> LSB als Pixel ausgeben
      b >>= 1;                                                        // nchste Bit-Position an LSB
    }
  }
  posx += 12;                                                         // X-Position fr nchstes Zeichen
}

void led_digit3(uint8_t d)                                            // Eine Kleinziffer 5x8 (Sekunden & Tag) in die LED-Matrix schreiben
{                                                                     // d: auszugebende Ziffer (0-9)
  uint16_t offset;                                                    // Offset in der Zeichensatz-Tabelle
  uint8_t x;                                                          // Zhler fr die Zeichenausgabe, X-Position
  uint8_t y;                                                          // Zhler fr die Zeichenausgabe, Y-Position
  uint8_t b;                                                          // Zwischenspeicher fr Bytes aus der Zeichensatz-Tabelle

  offset = d * 8;                                                     // Tabellen-Offset berechnen
  for (y = 0; y < 8; y ++) {                                          // 8 Zeilenbytes ausgeben
    b = pgm_read_byte(&digit_3_array[offset ++]);                     // Zeilenbyte holen
    for (x = 5; x > 0; x--) {                                         // Zeilenbyte bitweise ausgeben
      if (b & 1) matrix_pixel(posx + x - 1, posy + y);                // wenn Pixel gesetzt werden muss -> LSB als Pixel ausgeben
      b >>= 1;                                                        // nchste Bit-Position an LSB
    }
  }
  posx += 6;                                                          // X-Position fr nchstes Zeichen
}

uint8_t chrw(uint8_t c)                                               // Die Pixelbreite eines Zeichens ermitteln
{                                                                     // c: Zeichen
  uint16_t offset;                                                    // Offset in der Zeichensatz-Tabelle
  uint8_t charw = 0;                                                  // Pixelbreite des Zeichens zunchst auf 0 setzen

  if (c == 0) c = 127;                                                // wenn Nullzeichen -> durch Endezeichen ersetzen
  if (c >= 32) {                                                      // wenn zulssiger Zeichencode
    offset = (c - 32) * 12;                                           // Tabellen-Offset berechnen
    charw = pgm_read_byte(&charset_normal_array[offset]);             // Zeichenbreite holen
    if (charw > 8) charw = 8;                                         // Zeichenbreite auf 8 begrenzen
  }
  return charw;                                                       // Pixelbreite zurckgeben
}

uint8_t chrwr(uint8_t c)                                              // Die Pixelbreite eines reduzierten Zeichens ermitteln
{                                                                     // c: Zeichen
  uint16_t offset;                                                    // Offset in der Zeichensatz-Tabelle
  uint8_t charw = 0;                                                  // Pixelbreite des Zeichens zunchst auf 0 setzen

  if (c == 0) c = 127;                                                // wenn Nullzeichen -> durch Endezeichen ersetzen
  if (c >= 32) {                                                      // wenn zulssiger Zeichencode
    offset = (c - 32) * 10;                                           // Tabellen-Offset berechnen
    charw = pgm_read_byte(&charset_reduced_array[offset]);            // Zeichenbreite holen
    if (charw > 8) charw = 8;                                         // Zeichenbreite auf 8 begrenzen
  }
  return charw;                                                       // Pixelbreite zurckgeben
}

uint8_t strw(uint8_t strnr)                                           // Die Pixelbreite eines Strings im Flash-Speicher ermitteln
{                                                                     // strnr: String-Nummer im Flash-Speicher
  uint8_t i = 0;                                                      // Zeichenzhler
  PGM_P p;                                                            // Pointer auf Stringdaten im Flash
  uint8_t pw = 0;                                                     // Speicher fr Pixelbreite

  if (lang) p = (PGM_P) pgm_read_word(&stre[strnr]);                  // String-Adresse ermitteln (wenn Englisch gewhlt)
  else p = (PGM_P) pgm_read_word(&strd[strnr]);                       // String-Adresse ermitteln (wenn Deutsch gewhlt)
  strcpy_P(strbuf, p);                                                // String in den RAM-Puffer kopieren
  while (strbuf[i] > 0) {                                             // Schleife bis zum Endezeichen (0)
    pw += chrw(strbuf[i ++]) + 1;                                     // Pixelbreite eines Zeichens und einen Zwischenraum addieren
  }
  if (pw > 0) pw--;                                                   // wenn Pixelbreite > 0 -> korrigieren
  return pw;                                                          // Pixelbreite zurckgeben
}

uint8_t strwr2(uint8_t strnr)                                         // Die Pixelbreite eines reduzierten Strings im Flash-Speicher ermitteln (2 Zeichen fr Wochentag)
{                                                                     // strnr: String-Nummer im Flash-Speicher
  uint8_t i = 0;                                                      // Zeichenzhler
  PGM_P p;                                                            // Pointer auf Stringdaten im Flash
  uint8_t pw = 0;                                                     // Speicher fr Pixelbreite

  if (lang) p = (PGM_P) pgm_read_word(&stre[strnr]);                  // String-Adresse ermitteln (wenn Englisch gewhlt)
  else p = (PGM_P) pgm_read_word(&strd[strnr]);                       // String-Adresse ermitteln (wenn Deutsch gewhlt)
  strcpy_P(strbuf, p);                                                // String in den RAM-Puffer kopieren
  while ((strbuf[i] > 0) && (i < 2)) {                                // Schleife bis zum Endezeichen (0) oder bis zum zweiten Zeichen
    pw += chrwr(strbuf[i ++]) + 1;                                    // Pixelbreite eines reduzierten Zeichens und einen Zwischenraum addieren
  }
  if (pw > 0) pw--;                                                   // wenn Pixelbreite > 0 -> korrigieren
  return pw;                                                          // Pixelbreite zurckgeben
}

uint8_t symbolsw(uint8_t num)                                         // Die Pixelbreite eines speziellen Symbols ermitteln
{                                                                     // num: Nummer des Symbols (0-6)
  int8_t i;                                                           // Zhlervariable
  uint8_t w;                                                          // Symbolbreite
  PGM_P p;                                                            // Zeiger auf Symbol-Speicheradresse

  p = (PGM_P) &symbols;                                               // Symbol-Startadresse setzen
  for (i = 0; i < num; i ++) {                                        // Schleife zur Ermittlung der Startadresse des gewnschten Symbols
    w = pgm_read_byte(p ++);                                          // Symbolbreite holen
    if (w % 2) w ++;                                                  // wenn Symbolbreite ungerade -> auf nchsten geraden Wert setzen
    p += (w / 2) * 11;                                                // Symbol-Startadresse auf nchstes Symbol setzen
  }
  return pgm_read_byte(p);                                            // Pixelbreite zurckgeben
}

uint8_t senw(uint8_t pos)                                             // Die Pixelbreite eines Sensorfeld-Elements ermitteln
{                                                                     // pos: Feldposition
  uint8_t i = 0;                                                      // Zeichenzhler
  uint8_t pw = 0;                                                     // Speicher fr Pixelbreite
  char c;                                                             // Zwischenspeicher fr Windpfeil- oder Icon-Kennung

  while (i < 12) {                                                    // Schleife fr 12 Zeichen
    if (senarray[pos][i] > 0) {                                       // wenn kein Nullzeichen
      if ((senarray[pos][i] == '$') && (i < 11)) {                    // wenn ein Windpfeil oder Icon ausgegeben werden soll und noch nicht das letzte Zeichen erreicht ist
        i ++;                                                         // Zhler auf nchstes Zeichen
        c = senarray[pos][i];                                         // Windpfeil- oder Icon-Kennung holen
        if (c > 0) {                                                  // wenn kein Nullzeichen
          i ++;                                                       // Zhler auf nchstes Zeichen
          if ((c >= 'A') && (c <= 'S')) pw += 25;                     // wenn Icon-Kennung Zwischen A und S -> Icon-Breite und Zwischenraum addieren
          if ((c >= '0') && (c <= '8')) pw += chrw(c - '0' + 144) + 1; // wenn Windpfeil-Kennung zwischen 0 und 8 -> Windpfeil-Breite und Zwischenraum addieren
          if ((c >= 'T') && (c <= 'Z')) pw += symbolsw(c - 'T') + 1;  // wenn spezielles Symbol -> Symbol-Breite addieren
        }
      }
      else pw += chrw(senarray[pos][i ++]) + 1;                       // wenn kein Windpfeil oder Icon ausgegeben werden soll -> Pixelbreite eines Zeichens und einen Zwischenraum addieren
    }
    else break;                                                       // wenn Nullzeichen -> Abbruch
  }
  if (pw > 0) pw--;                                                   // wenn Pixelbreite > 0 -> letzten Zwischenraum entfernen
  return pw;                                                          // Pixelbreite zurckgeben
}

uint8_t almw(uint8_t pos)                                             // Die Pixelbreite eines Alarmfeld-Elements ermitteln
{                                                                     // pos: Feldposition
  uint8_t i = 0;                                                      // Zeichenzhler
  uint8_t pw = 0;                                                     // Speicher fr Pixelbreite

  while (i < 12) {                                                    // Schleife fr 12 Zeichen
    if (almarray[pos][i] > 0) pw += chrw(almarray[pos][i ++]) + 1;    // wenn kein Nullzeichen -> Pixelbreite eines Zeichens und einen Zwischenraum addieren
    else break;                                                       // wenn Nullzeichen -> Abbruch
  }
  if (pw > 0) pw--;                                                   // wenn Pixelbreite > 0 -> korrigieren
  return pw;                                                          // Pixelbreite zurckgeben
}

void led_chr(char c)                                                  // Ein ASCII-Zeichen in die LED-Matrix schreiben
{                                                                     // c: auszugebendes Zeichen (ASCII)
  uint16_t offset;                                                    // Offset in der Zeichensatz-Tabelle
  uint8_t cwidth;                                                     // Pixelbreite des Zeichens
  uint8_t x;                                                          // Zhler fr die Zeichenausgabe, X-Position
  uint8_t y;                                                          // Zhler fr die Zeichenausgabe, Y-Position
  uint8_t b;                                                          // Zwischenspeicher fr Bytes aus der Zeichensatz-Tabelle
  uint8_t c1, c2;                                                     // Zwischenspeicher fr Farbkomponenten

  if (c == 0) c = 127;                                                // wenn Nullzeichen -> durch Endezeichen ersetzen
  if (c >= 32) {                                                      // Ausgabe nur bei zulssigem Zeichencode
    c1 = color1;                                                      // Farbkomponente 1 sichern
    c2 = color2;                                                      // Farbkomponente 2 sichern
    offset = (c - 32) * 12;                                           // Tabellen-Offset berechnen
    cwidth = pgm_read_byte(&charset_normal_array[offset]);            // Zeichenbreite holen
    if (cwidth > 8) cwidth = 8;                                       // Zeichenbreite auf 8 begrenzen
    if (undrline) {                                                   // wenn Unterstreichung eingeschaltet
      color_date();                                                   // Farbe fr Datum einstellen
      matrix_pixel(posx - 1, posy + 10);                              // ein Pixel links neben dem Zeichen setzen
      matrix_pixel(posx + cwidth, posy + 10);                         // ein Pixel rechts neben dem Zeichen setzen
      color1 = c1;                                                    // Farbkomponente 1 wiederherstellen
      color2 = c2;                                                    // Farbkomponente 2 wiederherstellen
    }
    for (y = 0; y < 11; y ++) {                                       // 11 Zeilenbytes ausgeben
      b = pgm_read_byte(&charset_normal_array[offset + y + 1]);       // Zeilenbyte holen
      for (x = cwidth; x > 0; x--) {                                  // Zeilenbyte bitweise ausgeben
        if ((y < 10) && (b & 1))                                      // wenn Pixelzeile 0-9 und Bit gesetzt
          matrix_pixel(posx + x - 1, posy + y);                       // Pixel normal setzen
        if (y == 10) {                                                // wenn Pixelzeile 10
          if (b & 1) matrix_pixel(posx + x - 1, posy + y);            // wenn Bit gesetzt -> Pixel normal setzen
          else {                                                      // wenn kein Bit gesetzt
            if (undrline) {                                           // wenn Unterstreichung eingeschaltet
              color_date();                                           // Farbe fr Datum einstellen
              matrix_pixel(posx + x - 1, posy + y);                   // Unterstreichungs-Pixel setzen
              color1 = c1;                                            // Farbkomponente 1 wiederherstellen
              color2 = c2;                                            // Farbkomponente 2 wiederherstellen
            }
          }
        }
        b >>= 1;                                                      // nchstes Bit an Ausgabe-Position
      }
    }
    posx += cwidth + 1;                                               // X-Position fr nchstes Zeichen
  }
}

void led_chrr(char c)                                                 // Ein reduziertes ASCII-Zeichen in die LED-Matrix schreiben
{                                                                     // c: auszugebendes Zeichen (ASCII)
  uint16_t offset;                                                    // Offset in der Zeichensatz-Tabelle
  uint8_t cwidth;                                                     // Pixelbreite des Zeichens
  uint8_t x;                                                          // Zhler fr die Zeichenausgabe, X-Position
  uint8_t y;                                                          // Zhler fr die Zeichenausgabe, Y-Position
  uint8_t b;                                                          // Zwischenspeicher fr Bytes aus der Zeichensatz-Tabelle

  if (c == 0) c = 127;                                                // wenn Nullzeichen -> durch Endezeichen ersetzen
  if (c >= 32) {                                                      // Ausgabe nur bei zulssigem Zeichencode
    offset = (c - 32) * 10;                                           // Tabellen-Offset berechnen
    cwidth = pgm_read_byte(&charset_reduced_array[offset]);           // Zeichenbreite holen
    if (cwidth > 8) cwidth = 8;                                       // Zeichenbreite auf 8 begrenzen
    for (y = 0; y < 9; y ++) {                                        // 9 Zeilenbytes ausgeben
      b = pgm_read_byte(&charset_reduced_array[offset + y + 1]);      // Zeilenbyte holen
      for (x = cwidth; x > 0; x--) {                                  // Zeilenbyte bitweise ausgeben
        if (b & 1) matrix_pixel(posx + x - 1, posy + y);              // wenn Bit gesetzt -> Pixel setzen
        b >>= 1;                                                      // nchstes Bit an Ausgabe-Position
      }
    }
    posx += cwidth + 1;                                               // X-Position fr nchstes Zeichen
  }
}

void led_icon(uint8_t n)                                              // Ein Icon als Grafik in die LED-Matrix schreiben
{                                                                     // n: auszugebende Icon-Nummer
  uint8_t x;                                                          // Zhler fr die Zeichenausgabe, X-Position
  uint8_t y;                                                          // Zhler fr die Zeichenausgabe, Y-Position
  uint8_t b;                                                          // Zwischenspeicher fr Bytes aus der Icon-Liste
  uint8_t c;                                                          // Zwischenspeicher fr Farbwert
  uint8_t c1, c2;                                                     // Zwischenspeicher fr Farbkomponenten
  PGM_P p;                                                            // Zeiger auf Icon-Speicheradresse

  c1 = color1;                                                        // Farbkomponente 1 sichern
  c2 = color2;                                                        // Farbkomponente 2 sichern
  p = (PGM_P) &iconarray + n * 132;                                   // Icon-Adresse berechnen
  for (y = 0; y < 11; y ++) {                                         // 11 Zeilen ausgeben
    for (x = 0; x < 12; x ++) {                                       // 12 Bytes (24 Pixel) ausgeben
      b = pgm_read_byte(p ++);                                        // Farbbyte holen
      c = b / 16;                                                     // Farbwert fr linkes Pixel ermitteln
      if (c > 0) {                                                    // wenn Farbwert > 0
        matrix_color(c);                                              // Farbe setzen
        matrix_pixel(posx + x * 2, posy + y);                         // Pixel ausgeben
      }
      c = b & 0x0f;                                                   // Farbwert fr rechtes Pixel ermitteln
      if (c > 0) {                                                    // wenn Farbwert > 0
        matrix_color(c);                                              // Farbe setzen
        matrix_pixel(posx + x * 2 + 1, posy + y);                     // Pixel ausgeben
      }
    }
  }
  posx += 25;                                                         // X-Position fr nchstes Zeichen
  color1 = c1;                                                        // Farbkomponente 1 wiederherstellen
  color2 = c2;                                                        // Farbkomponente 2 wiederherstellen
}

void led_symbols(uint8_t num)                                         // Ein spezielles Symbol in die LED-Matrix schreiben
{
  int8_t i;                                                           // Zhlervariable
  uint8_t w;                                                          // Zeichenbreite in Pixel
  uint8_t x;                                                          // Zhler fr die Zeichenausgabe, X-Position
  uint8_t y;                                                          // Zhler fr die Zeichenausgabe, Y-Byte-Position
  uint8_t e;                                                          // Endwert fr Y-Byte-Position
  uint8_t b;                                                          // Zwischenspeicher fr Symbol-Bytes
  uint8_t c;                                                          // Zwischenspeicher fr Farbwert
  uint8_t c1, c2;                                                     // Zwischenspeicher fr Farbkomponenten
  PGM_P p;                                                            // Zeiger auf Icon-Speicheradresse

  c1 = color1;                                                        // Farbkomponente 1 sichern
  c2 = color2;                                                        // Farbkomponente 2 sichern
  p = (PGM_P) &symbols;                                               // Symbol-Startadresse setzen
  for (i = 0; i < num; i ++) {                                        // Schleife zur Ermittlung der Startadresse des gewnschten Symbols
    w = pgm_read_byte(p ++);                                          // Symbolbreite holen
    if (w % 2) w ++;                                                  // wenn Symbolbreite ungerade -> auf nchsten geraden Wert setzen
    p += (w / 2) * 11;                                                // Symbol-Startadresse auf nchstes Symbol setzen
  }
  w = pgm_read_byte(p ++);                                            // Symbolbreite holen
  for (y = 0; y < 11; y ++) {                                         // 11 Zeilen ausgeben
    if (w % 2) e = (w + 1) / 2;                                       // wenn Symbolbreite ungerade -> Y-Endwert um ein "halbes" Byte ergnzen
    else e = w / 2;                                                   // sonst entspricht der Y-Endwert der halben Symbolbreite
    for (x = 0; x < e; x ++) {                                        // ermittelte Anzahl Bytes ausgeben
      b = pgm_read_byte(p ++);                                        // Farbbyte holen
      c = b / 16;                                                     // Farbwert fr linkes Pixel ermitteln
      if (c > 0) {                                                    // wenn Farbwert > 0
        matrix_color(c);                                              // Farbe setzen
        matrix_pixel(posx + x * 2, posy + y);                         // Pixel ausgeben
      }
        if (x * 2 + 1 < w) {                                          // wenn ungerade Symbolbreite noch nicht erreicht
        c = b & 0x0f;                                                 // Farbwert fr rechtes Pixel ermitteln
        if (c > 0) {                                                  // wenn Farbwert > 0
          matrix_color(c);                                            // Farbe setzen
          matrix_pixel(posx + x * 2 + 1, posy + y);                   // Pixel ausgeben
        }
      }
    }
  }
  posx += w + 1;                                                      // X-Position fr nchstes Zeichen
  color1 = c1;                                                        // Farbkomponente 1 wiederherstellen
  color2 = c2;                                                        // Farbkomponente 2 wiederherstellen
}

void copy_str(uint8_t strnr)                                          // Einen String aus dem Flash-Speicher in den Stringpuffer kopieren
{                                                                     // strnr: String-Nummer im Flash-Speicher
  PGM_P p;                                                            // Speicheradresse auf Stringdaten im Flash

  if (lang) p = (PGM_P) pgm_read_word(&stre[strnr]);                  // String-Adresse ermitteln (wenn Englisch gewhlt)
  else p = (PGM_P) pgm_read_word(&strd[strnr]);                       // String-Adresse ermitteln (wenn Deutsch gewhlt)
  strcpy_P(strbuf, p);                                                // String in den RAM-Puffer kopieren
}

void send_str(uint8_t strnr)                                          // Einen String aus dem Flash-Speicher ber die serielle Schnittstelle senden
{                                                                     // strnr: String-Nummer
  PGM_P p;                                                            // Speicheradresse auf Stringdaten im Flash

  if (lang) p = (PGM_P) pgm_read_word(&stre[strnr]);                  // String-Adresse ermitteln (wenn Englisch gewhlt)
  else p = (PGM_P) pgm_read_word(&strd[strnr]);                       // String-Adresse ermitteln (wenn Deutsch gewhlt)
  while (pgm_read_byte(p) > 0) {                                      // Schleife bis zum Endezeichen
    while (!(UCSR1A & _BV(UDRE1)));                                   // warten bis Sendepuffer leer ist
    UDR1 = pgm_read_byte(p);                                          // Zeichen in Sendepuffer legen
    p ++;                                                             // Zhler erhhen
  }
}

void led_str(uint8_t strnr)                                           // Einen String aus dem Flash-Speicher in die LED-Matrix schreiben
{                                                                     // strnr: String-Nummer im Flash-Speicher
  uint8_t i = 0;                                                      // Zeichenzhler auf erstes Zeichen setzen

  copy_str(strnr);                                                    // String in RAM-Puffer kopieren
  while (strbuf[i] > 0) {                                             // Schleife bis zum Endezeichen (0)
    led_chr(strbuf[i ++]);                                            // ein Zeichen ausgeben
  }
}

void led_strr2(uint8_t strnr)                                         // Einen reduzierten String aus dem Flash-Speicher in die LED-Matrix schreiben (2 Zeichen fr Wochentag)
{                                                                     // strnr: String-Nummer im Flash-Speicher
  uint8_t i = 0;                                                      // Zeichenzhler auf erstes Zeichen setzen

  copy_str(strnr);                                                    // String in RAM-Puffer kopieren
  while ((strbuf[i] > 0) && (i < 2)) {                                // Schleife bis zum Endezeichen (0) oder bis zum zweiten Zeichen
    led_chrr(strbuf[i ++]);                                           // ein reduziertes Zeichen ausgeben
  }
}

void led_sen(uint8_t pos)                                             // Ein Sensorfeld-Element ausgeben
{                                                                     // pos: Feldposition
  uint8_t i = 0;                                                      // Zeichenzhler auf erstes Zeichen setzen
  char c;                                                             // Zwischenspeicher fr Windpfeil- oder Icon-Kennung

  while (i < 12) {                                                    // Schleife fr 12 Zeichen
    if (senarray[pos][i] > 0) {                                       // wenn kein Nullzeichen
      if ((senarray[pos][i] == '$') && (i < 11)) {                    // wenn ein Windpfeil oder Icon ausgegeben werden soll und noch nicht das letzte Zeichen erreicht ist
        i ++;                                                         // Zhler auf nchstes Zeichen
        c = senarray[pos][i];                                         // Windpfeil- oder Icon-Kennung holen
        if (c > 0) {                                                  // wenn kein Nullzeichen
          i ++;                                                       // Zhler auf nchstes Zeichen
          if ((c >= 'A') && (c <= 'S')) led_icon(c - 'A');            // wenn Icon-Kennung Zwischen A und S -> Icon ausgeben
          if ((c >= '0') && (c <= '8')) led_chr(c - '0' + 144);       // wenn Windpfeil-Kennung zwischen 0 und 8 -> Windpfeil ausgeben
          if ((c >= 'T') && (c <= 'Z')) led_symbols(c - 'T');         // wenn spezielles Symbol -> Symbol ausgeben
        }
      }
      else led_chr(senarray[pos][i ++]);                              // wenn kein Icon ausgegeben werden soll -> Zeichen normal ausgeben
    }
    else break;                                                       // wenn Nullzeichen -> Abbruch
  }
}

void led_alm(uint8_t pos)                                             // Ein Alarmfeld-Element ausgeben
{                                                                     // pos: Feldposition
  uint8_t i = 0;                                                      // Zeichenzhler auf erstes Zeichen setzen

  while (i < 12) {                                                    // Schleife fr 12 Zeichen
    if (almarray[pos][i] > 0) led_chr(almarray[pos][i ++]);           // wenn kein Nullzeichen -> Zeichen ausgeben
    else break;                                                       // wenn Nullzeichen -> Abbruch
  }
}

uint8_t calc_wday(uint8_t d, uint8_t m, uint8_t yh, int8_t y)         // Wochentags-Berechnung nach Gauss (nach http://www.paswiki.de/index.php/Ewiger_Kalender, existiert leider nicht mehr)
{                                                                     // Ergebnis: Wochentag (1-7)
                                                                      // d: Tag, m: Monat, yh: Jahrhundert, y: Jahr
  int8_t w;                                                           // Wochtag

  if (m <= 2) {                                                       // wenn Monat Januar oder Februar
    m += 12;                                                          // Monat ans Jahresende verschieben
    y--;                                                              // Jahr vermindern
    if (y < 0) {                                                      // wenn bertrag
      y = 99;                                                         // Jahr auf 99 setzen
      yh--;                                                           // Jahrhundert vermindern
    }
  }
  w = (d + (13 * (m + 1)) / 5 + y + y / 4 + yh / 4 - 2 * yh) % 7;     // Wochtagsformel, kleinster Wert entspricht Samstag
  if (w < 0) w += 7;                                                  // falls Ergebnis negativ war
  w--;                                                                // Wochentag korrigieren (kleinster Wert ist Montag)
  if (w < 1) w += 7;                                                  // und in den Bereich 1-7 legen
  return w;                                                           // Wochentag zurckgeben
}

uint8_t chk_summtime(void)                                            // Prfen, ob aktuelles Datum in der Sommerzeit liegt
{                                                                     // Ergebnis: 0 = Normalzeit, 1 = Sommerzeit
  uint8_t stime = 0;                                                  // Anfangswert auf Normalzeit setzen

  if ((month > 3) && (month < 10)) stime = 1;                         // wenn Monat April bis September -> Sommerzeit setzen
  if ((month == 3) && (day > 24)) {                                   // wenn Datum im Zeitraum vom 25. bis 31. Mrz
    if (calc_wday(day, month, c_yearh, year) == 7) stime = 1;         // wenn Sonntag -> Sommerzeit setzen (Uhrzeit wird hier nicht bercksichtigt)
    else if ((day - calc_wday(day, month, c_yearh, year)) > 24) stime = 1; // wenn nicht Sonntag -> wenn Tag nach der Umstellung -> Sommerzeit setzen
  }
  if ((month == 10) && (day > 24)) {                                  // wenn Datum im Zeitraum vom 25. bis 31. Oktober
    if (calc_wday(day, month, c_yearh, year) == 7) stime = 0;         // wenn Sonntag -> Normalzeit setzen (Uhrzeit wird hier nicht bercksichtigt)
    else if ((day - calc_wday(day, month, c_yearh, year)) < 25) stime = 1; // wenn nicht Sonntag -> wenn Tag vor der Umstellung -> Sommerzeit setzen
  }
  return stime;                                                       // Sommerzeit-Status zurckgeben
}

void check_date(void)                                                 // Datum (Tag) und Wochentag korrigieren
{
  if ((year % 4) == 0) {                                              // wenn Schaltjahr
    if (day > pgm_read_byte(&daytab2[month - 1]))                     // wenn Tag > Schaltjahr-Tabelle
      day = pgm_read_byte(&daytab2[month - 1]);                       // auf letzten mglichen Tag setzen
  }
  else {                                                              // wenn kein Schaltjahr
    if (day > pgm_read_byte(&daytab1[month - 1]))                     // wenn Tag > Normaljahr-Tabelle
      day = pgm_read_byte(&daytab1[month - 1]);                       // auf letzten mglichen Tag setzen
  }
  wday = calc_wday(day, month, c_yearh, year);                        // Wochentag berechnen
}

char next_char(char c)                                                // Nchstes Zeichen aus der Zeichenliste ermitteln
{
  uint8_t i;                                                          // Zhlervariable
  uint8_t p = 255;                                                    // Zeichenposition auf ungltigen Wert setzen

  if (c == 0) c = 127;                                                // wenn Nullzeichen -> zu darstellbarem Endezeichen wandeln
  for (i = 0; i < sizeof(charlist); i ++) {                           // Zeichen in der Liste suchen
    if (pgm_read_byte(&charlist[i]) == c) p = i;                      // wenn gefunden -> Position setzen
  }
  if (p < 255) p ++;                                                  // wenn Position gefunden -> auf nchste Position setzen
  else p = 0;                                                         // sonst auf Anfang setzen
  if (p >= sizeof(charlist)) p = 0;                                   // wenn Listenende erreicht -> Position auf Anfang setzen
  c = pgm_read_byte(&charlist[p]);                                    // nchstes Zeichen aus der Liste holen
  if (c == 127) c = 0;                                                // wenn Endezeichen -> zu Nullzeichen wandeln
  return c;
}

char utf8_conv(uint8_t c1, uint8_t c2)                                // Ein UTF-8 Zeichen (2 Byte) in ein internes Zeichen konvertieren
{
  uint8_t i;                                                          // Zhlervariable
  uint8_t cc = ' ';                                                   // konvertiertes Zeichen auf Leerzeichen voreinstellen

  for (i = 0; i < sizeof(utf8tab); i += 3) {                          // UTF-8-Tabelle in Schritten von 3 Bytes abfragen
    if (cc == ' ') {                                                  // wenn noch kein Zeichen gefunden
      if (pgm_read_byte(&utf8tab[i]) == c1) {                         // wenn erstes Zeichen gefunden
        if (pgm_read_byte(&utf8tab[i + 1]) == c2) {                   // wenn zweites Zeichen gefunden
          cc = pgm_read_byte(&utf8tab[i + 2]);                        // internes Zeichen lesen
        }
      }
    }
  }
  return cc;                                                          // konvertiertes Zeichen zurckgeben
}

char utf8_conv3(uint8_t c1, uint8_t c2, uint8_t c3)                   // Ein UTF-8 Zeichen (3 Byte) in ein internes Zeichen konvertieren
{
  uint8_t i;                                                          // Zhlervariable
  uint8_t cc = ' ';                                                   // konvertiertes Zeichen auf Leerzeichen voreinstellen

  for (i = 0; i < sizeof(utf8tab3); i += 4) {                         // UTF-8-Tabelle in Schritten von 4 Bytes abfragen
    if (cc == ' ') {                                                  // wenn noch kein Zeichen gefunden
      if (pgm_read_byte(&utf8tab3[i]) == c1) {                        // wenn erstes Zeichen gefunden
        if (pgm_read_byte(&utf8tab3[i + 1]) == c2) {                  // wenn zweites Zeichen gefunden
          if (pgm_read_byte(&utf8tab3[i + 2]) == c3) {                // wenn drittes Zeichehn gefunden
            cc = pgm_read_byte(&utf8tab3[i + 3]);                     // internes Zeichen lesen
          }          
        }
      }
    }
  }
  return cc;                                                          // konvertiertes Zeichen zurckgeben
}

void sensor_sm(void)                                                  // Neuen Datensatz vom Sensormodul ins Sensorfeld eintragen bzw. Datensatz aktualisieren
{
  uint8_t i;                                                          // Zhlervariable
  uint8_t j = 0;                                                      // Zeigervariable auf Stringanfang setzen
  uint8_t m;                                                          // Merker fr Sensorsuche
  uint8_t s = 255;                                                    // Sensornummer auf ungltigen Wert setzen

  switch (rxdata[0]) {                                                // Sensorkennung auswerten und Sensornummer ermitteln
    case '1' ... '8':                                                 // falls Sensorkennung 1-8
      s = rxdata[0] - 49;                                             // in logische Sensornummer 0-7 umrechnen
      break;
    case 'a' ... 's':                                                 // falls Sensorkennung a-s
      s = rxdata[0] - 89;                                             // in logische Sensornummer 8-26 umrechnen
      break;
    case 'w' ... 'z':                                                 // falls Sensorkennung w-z
      s = rxdata[0] - 92;                                             // in logische Sensornummer 27-30 umrechnen
      break;
  }
  if (s < 31) {                                                       // wenn gltige Sensornummer (0-30)
    m = 255;                                                          // Merker auf ungltigen Wert setzen
    if (senfil) {                                                     // wenn Sensorfilter eingeschaltet ist
      for (i = 0; i < 8; i ++) {                                      // 8 Sensornummern durchsuchen
        if (sennum[i] == (s + 1)) m = i;                              // wenn Sensornummer gefunden -> Merker auf Sensorplatz setzen
      }
      if (m < 8) {                                                    // wenn Sensor gefunden
        while (j < 5) {                                               // wiederholen bis alle 5 Zeichen bearbeitet
          if (sennam[m][j] > 0) {                                     // wenn kein Nullzeichen
            senarray[s][j] = sennam[m][j];                            // ein Zeichen des Namens ins Sensorfeld kopieren
            j ++;                                                     // nchste Zeichenposition
          }
          else break;                                                 // wenn Nullzeichen -> Abbruch
        }
        if (j < 5) senarray[s][j ++] = ' ';                           // wenn Sensorname krzer als 5 Zeichen -> Leerzeichen einfgen
        else {                                                        // wenn Sensorname 5 Zeichen lang
          if (rxdata[2] == ' ') senarray[s][j ++] = ' ';              // Leerzeichen nur dann einfgen, wenn erstes Zeichen im Sensorwert leer ist
        }
      }
    }
    else {                                                            // wenn Sensorfilter ausgeschaltet ist
      senarray[s][j ++] = 'S';                                        // 'S' als 1. Zeichen setzen
      senarray[s][j ++] = (s + 1) / 10 + '0';                         // reale Sensornummer-Zehner als 2. Zeichen setzen
      senarray[s][j ++] = (s + 1) % 10 + '0';                         // reale Sensornummer-Einer als 3. Zeichen setzen
      senarray[s][j ++] = ':';                                        // Doppelpunkt als 4. Zeichen setzen
      senarray[s][j ++] = ' ';                                        // Leerzeichen als 5. Zeichen setzen
      m = 0;                                                          // Merker auf gltigen Wert setzen, damit Sensorwert und Einheit ergnzt wird
    }
    if (m < 8) {                                                      // wenn Merker gltig
      for (i = 2; i < rxpos; i ++) {                                  // restliche Zeichen der Empfangsdaten kopieren
        if (rxdata[i] > ' ') {                                        // wenn kein Steuer- oder Leerzeichen
          senarray[s][j ++] = rxdata[i];                              // Zeichen kopieren und Zeiger erhhen
        }
      }
      switch (s) {                                                    // Sensornummer auswerten und Einheit ermitteln (C, hPa, %)
        case 0 ... 26:                                                // falls logische Sensornummer 0-26
          senarray[s][j ++] = 138;                                    // "C" ergnzen
          senarray[s][j ++] = 'C';
          break;
        case 27:                                                      // falls logische Sensornummer 27
          senarray[s][j ++] = 'h';                                    // "hP" ergnzen
          senarray[s][j ++] = 'P';
          if (senarray[s][j - 6] == ' ')                              // wenn Luftdruck < 1000 hPa
          senarray[s][j ++] = 'a';                                    // 'a' ergnzen
          break;
        case 28 ... 30:                                               // falls logische Sensornummer 28-30
          senarray[s][j ++] = '%';                                    // '%' ergnzen
          break;
      }
      while (j < 12) {senarray[s][j ++] = 0;}                         // mit Nullzeichen auffllen
      sentout[s] = c_setout / chtime;                                 // Timeout-Zhler neu setzen
    }
  }
}

void sensor_wm(void)                                                  // Neuen Datensatz vom Web-Modul ins Sensorfeld eintragen bzw. Datensatz aktualisieren
{
  uint8_t i = 3;                                                      // Zeigervariable auf Datenbereich der Empfangsdaten setzen
  uint8_t j = 0;                                                      // Zeigervariable auf Stringanfang setzen
  uint8_t s = 255;                                                    // Sensornummer auf ungltigen Wert setzen
  char c, c1;                                                         // Zeichen-Zwischenspeicher

  if ((rxdata[1] >= '0') && (rxdata[1] <= '9')) {                     // wenn zweite Stelle Ziffer zwischen 0 und 9
    s = (rxdata[0] - '0') * 10 + rxdata[1] - '0' - 1;                 // Sensornummer ermitteln und in logische Nummer wandeln (0-30)
  }
  if (s < 31) {                                                       // wenn gltige Sensornummer (0-30)
    if (rxpos == 3) {                                                 // wenn keine weiteren Zeichen nach dem Doppelpunkt
      sentout[s] = 0;                                                 // Timeout-Zhler lschen, Sensorfeld wird im nchsten Zyklus gelscht
    }
    else {                                                            // wenn weitere Zeichen nach dem Doppelpunkt folgen
      while (i < rxpos) {                                             // solange noch Zeichen im Puffer sind
        if (j < 12) {                                                 // wenn noch Platz im Sensorfeld
          c = rxdata[i ++];                                           // Zeichen aus Empfangspuffer holen und Zeiger erhhen
          if ((c > 127) && (i < rxpos)) {                             // wenn UTF-8-Zeichen und noch ein weiteres Zeichen im Puffer
            if (c < 0xE0) c = utf8_conv(c, rxdata[i ++]);             // wenn 2-Byte-Zeichen -> Zeichen konvertieren (2 Byte)
            else {                                                    // sonst
              c1 = rxdata[i ++];                                      // zweites Zeichen holen
              if (i < rxpos) {                                        // wenn noch weitere Zeichen im Puffer
                c = utf8_conv3(c, c1, rxdata[i ++]);                  // Zeichen konvertieren (3 Byte)
              }
            }
          }
          senarray[s][j ++] = c;                                      // Zeichen ins Sensorfeld legen
          if (j < 12) senarray[s][j] = 0;                             // wenn noch Platz im Sensorfeld -> Endezeichen ergnzen
        }
        else i ++;                                                    // wenn kein Platz mehr im Sensorfeld -> Zeichen im Puffer ignorieren
      }
      sentout[s] = c_setout / chtime;                                 // Timeout-Zhler neu setzen
    }    
  }
}

void alarm_sm(void)                                                   // Neuen Datensatz vom Sensormodul ins Alarmfeld eintragen bzw. Datensatz aktualisieren
{
  uint8_t j = 0;                                                      // Zeigervariable auf Stringanfang setzen
  uint8_t a = 0;                                                      // Alarmnummer

  if ((rxdata[0] >= 'A') && (rxdata[0] <= 'D')) {                     // wenn gltige Alarmnummer
    if ((rxdata[2] == '0') || (rxdata[2] == '1')) {                   // wenn gltiger Alarmwert
      a = rxdata[0] - 'A';                                            // logische Alarmnummer ermitteln
      almsnew[a] = rxdata[2] - '0';                                   // neuen Alarmwert speichern
      if (alarms) {                                                   // wenn Alarme des Sensormoduls signalisiert werden sollen
        if (almsnew[a] != almnor[a]) {                                // wenn Alarm aktiv ist
          almtout[a] = c_altout / chtime;                             // Timeout-Zhler neu setzen
          if (almsnew[a] != almsold[a]) {                             // wenn nderung des Alarmstatus
            while (j < 12) {                                          // wiederholen bis alle 12 Zeichen bearbeitet
              if (almnam[a][j] > 0) {                                 // wenn kein Nullzeichen
                almarray[a][j] = almnam[a][j];                        // ein Zeichen des Namens ins Alarmfeld kopieren
                j ++;                                                 // nchste Zeichenposition
              }
              else break;                                             // wenn Nullzeichen -> Abbruch
            }
            while (j < 12) {almarray[a][j ++] = 0;}                   // mit Nullzeichen auffllen
            almstat[a] = 1;                                           // Alarmstatus auf Aktiv setzen
            almhold[a] = 1;                                           // Halte-Flag setzen (Alarm vom Sensormodul muss immer quittiert werden)
            if (syncstat && almsnd && sounflag && !mute)              // wenn Uhr synchronisiert, akustische Alarme eingeschaltet, Soundzeit und keine Stummschaltung
              sndflags[a] = 1;                                        // Sound-Flag setzen, akustischen Alarm auslsen
            almcoun = a;                                              // Alarm-Zhler auf neuen Alarm setzen
            if (alm_flag == 0) alm_flag = 1;                          // wenn Alarm-Flag noch nicht gesetzt -> Alarm-Flag setzen
          }
        }
        else {                                                        // wenn Alarm inaktiv ist
          if (almsnew[a] != almsold[a]) {                             // wenn nderung des Alarmstatus
            almstat[a] = 0;                                           // Alarmstatus auf Inaktiv setzen
          }
        }
      }
      almsold[a] = almsnew[a];                                        // neuen Alarmstatus kopieren
    }
  }
}

void alarm_wm(void)                                                   // Neuen Datensatz vom Web-Modul ins Alarmfeld eintragen bzw. Datensatz aktualisieren
{
  uint8_t i = 6;                                                      // Zeigervariable auf Datenbereich der Empfangsdaten (Alarmname) setzen
  uint8_t j = 0;                                                      // Zeigervariable auf Stringanfang setzen
  uint8_t a = 0;                                                      // Alarmnummer
  char c, c1;                                                         // Zeichen-Zwischenspeicher

  if ((rxdata[1] >= '1') && (rxdata[1] <= '8')) {                     // wenn gltige Alarmnummer
    if ((rxdata[3] == '0') || (rxdata[3] == '1')) {                   // wenn gltiger Alarmwert
      a = rxdata[1] - '1';                                            // logische Alarmnummer ermitteln
      if ((rxdata[3] == '1') && (rxpos > 6)) {                        // wenn Alarm aktiv ist und alle Parameter vorhanden sind
        if ((rxdata[4] >= '0') && (rxdata[4] <= '6')) {               // wenn Alarmsound im Bereich 0-6
          if ((rxdata[5] == '0') || (rxdata[5] == '1')) {             // wenn gltiger Halte-Modus
            almhold[a] = rxdata[5] - '0';                             // Halte-Flag aus Datensatz bernehmen
            almtout[a] = c_altout / chtime;                           // Timeout-Zhler neu setzen
            while (i < rxpos) {                                       // solange noch Zeichen im Puffer sind
              if (j < 12) {                                           // wenn noch Platz im Alarmfeld
                c = rxdata[i ++];                                     // Zeichen aus Empfangspuffer holen und Zeiger erhhen
                if ((c > 127) && (i < rxpos)) {                       // wenn UTF-8-Zeichen und noch ein weiteres Zeichen im Puffer
                  if (c < 0xE0) c = utf8_conv(c, rxdata[i ++]);       // wenn 2-Byte-Zeichen -> Zeichen konvertieren (2 Byte)
                  else {                                              // sonst
                    c1 = rxdata[i ++];                                // zweites Zeichen holen
                    if (i < rxpos) {                                  // wenn noch weitere Zeichen im Puffer
                      c = utf8_conv3(c, c1, rxdata[i ++]);            // Zeichen konvertieren (3 Byte)
                    }
                  }
                }
                almarray[a][j ++] = c;                                // Zeichen ins Alarmfeld legen
                if (j < 12) almarray[a][j] = 0;                       // wenn noch Platz im Alarmfeld -> Endezeichen ergnzen
              }
              else i ++;                                              // wenn kein Platz mehr im Alarmfeld -> Zeichen im Puffer ignorieren
            }
            if (almstat[a] == 0) {                                    // wenn bisheriger Status inaktiv
              almstat[a] = 1;                                         // Status auf Aktiv setzen
              almcoun = a;                                            // Alarm-Zhler auf neuen Alarm setzen
              if (alm_flag == 0) alm_flag = 1;                        // wenn Alarm-Flag noch nicht gesetzt -> Alarm-Flag setzen
              if (syncstat && almsnd && sounflag && !mute) {          // wenn Uhr synchronisiert, akustische Alarme eingeschaltet, Soundzeit und keine Stummschaltung
                if (rxdata[4] > '0') {                                // wenn einer der 6 Alarm-Sounds ausgewhlt ist
                  sndflags[rxdata[4] - '1'] = 1;                      // Sound-Flag setzen, akustischen Alarm auslsen
                }
              }
            }
          }
        }
      }
      if ((rxdata[3] == '0') && (rxpos > 3)) {                        // wenn Alarm inaktiv ist und alle Parameter vorhanden sind
        if (almstat[a] == 1) {                                        // wenn bisheriger Status aktiv
          almstat[a] = 0;                                             // Alarmstatus auf Inaktiv setzen
          almtout[a] = c_altout / chtime;                             // Timeout-Zhler ein letztes mal neu setzen
        }
      }
    }
  }
}

void time_wm(void)                                                    // Zeitdatensatz vom Web-Modul prfen und Uhr stellen, z.B. "TM:20180313 17:11:47.18 1"
{
  uint8_t yh = 0;                                                     // Zwischenspeicher fr Jahrhundert
  uint8_t yl = 0;                                                     // Zwischenspeicher fr Jahr
  uint8_t mo = 0;                                                     // Zwischenspeicher fr Monat
  uint8_t d = 0;                                                      // Zwischenspeicher fr Tag
  uint8_t h = 0;                                                      // Zwischenspeicher fr Stunde
  uint8_t m = 0;                                                      // Zwischenspeicher fr Minute
  uint8_t s = 0;                                                      // Zwischenspeicher fr Sekunde
  uint8_t hs = 0;                                                     // Zwischenspeicher fr Hundertstel-Sekunden
  uint8_t tz = 0;                                                     // Zwischenspeicher fr Zeitzone
  uint8_t c = 0;                                                      // Prfzhler, wird bei jedem erfolgreichen Prfschritt erhht
  uint8_t i;                                                          // Zhlervariable

  for (i = 3; i < 11; i ++) {                                         // 8 Stellen des Datums prfen
    if ((rxdata[i] >= '0') && (rxdata[i] <= '9')) c ++;               // Ziffer 0-9? -> Prfzhler erhhen
  }
  if (rxdata[11] == ' ') c ++;                                        // Leerzeichen zwischen Datum und Zeit? -> Prfzhler erhhen
  for (i = 12; i < 14; i ++) {                                        // 2 Stellen der Stunde prfen
    if ((rxdata[i] >= '0') && (rxdata[i] <= '9')) c ++;               // Ziffer 0-9? -> Prfzhler erhhen
  }
  if (rxdata[14] == ':') c ++;                                        // Doppelpunkt zwischen Stunde und Minute? -> Prfzhler erhhen
  for (i = 15; i < 17; i ++) {                                        // 2 Stellen der Stunde prfen
    if ((rxdata[i] >= '0') && (rxdata[i] <= '9')) c ++;               // Ziffer 0-9? -> Prfzhler erhhen
  }
  if (rxdata[17] == ':') c ++;                                        // Doppelpunkt zwischen Minute und Sekunde? -> Prfzhler erhhen
  for (i = 18; i < 20; i ++) {                                        // 2 Stellen der Minute prfen
    if ((rxdata[i] >= '0') && (rxdata[i] <= '9')) c ++;               // Ziffer 0-9? -> Prfzhler erhhen
  }
  if (rxdata[20] == '.') c ++;                                        // Punkt zwischen Sekunde und Hundertstel? -> Prfzhler erhhen
  for (i = 21; i < 23; i ++) {                                        // 2 Stellen der Sekunde prfen
    if ((rxdata[i] >= '0') && (rxdata[i] <= '9')) c ++;               // Ziffer 0-9? -> Prfzhler erhhen
  }
  if (rxdata[23] == ' ') c ++;                                        // Leerzeichen zwischen Hundertstel und Zeitzone -> Prfzhler erhhen
  if ((rxdata[24] == '0') || (rxdata[24] == '1')) c ++;               // Zeitzone 0 oder 1? -> Prfzhler erhhen

  if (c == 22) {                                                      // wenn alle bisherigen Prfungen erfolgreich
    yh = (rxdata[3] - '0') * 10 + rxdata[4] - '0';                    // Jahrhundert ermitteln
    yl = (rxdata[5] - '0') * 10 + rxdata[6] - '0';                    // Jahr ermitteln
    mo = (rxdata[7] - '0') * 10 + rxdata[8] - '0';                    // Monat ermitteln
    d = (rxdata[9] - '0') * 10 + rxdata[10] - '0';                    // Tag ermitteln
    h = (rxdata[12] - '0') * 10 + rxdata[13] - '0';                   // Stunde ermitteln
    m = (rxdata[15] - '0') * 10 + rxdata[16] - '0';                   // Minute ermitteln
    s = (rxdata[18] - '0') * 10 + rxdata[19] - '0';                   // Sekunde ermitteln
    hs = (rxdata[21] - '0') * 10 + rxdata[22] - '0';                  // Hundertstel ermitteln
    tz = rxdata[24] - '0';                                            // Zeitzone ermitteln
    if (yh == 20) c ++;                                               // wenn Jahrhundert stimmt -> Prfzhler erhhen
    if ((mo >= 1) && (mo <= 12)) c ++;                                // wenn Monat im Bereich 1-12 -> Prfzhler erhhen
    if ((d >= 1) && (d <= 31)) c ++;                                  // wenn Tag im Bereich 1-31 -> Prfzhler erhhen
    if (h < 24) c ++;                                                 // wenn Stunde im Bereich 0-23 -> Prfzhler erhhen
    if (m < 60) c ++;                                                 // wenn Minute im Bereich 0-59 -> Prfzhler erhhen
    if (s < 60) c ++;                                                 // wenn Sekunde im Bereich 0-59 -> Prfzhler erhhen
  }

  if (c == 28) {                                                      // wenn alle Prfungen erfolgreich
    year = yl;                                                        // Jahr setzen
    month = mo;                                                       // Monat setzen
    day = d;                                                          // Tag setzen
    wday = calc_wday(day, month, c_yearh, year);                      // Wochentag berechnen
    if (timezone != tz) {                                             // wenn sich die Zeitzone gendert hat
      timezone = tz;                                                  // neue Zeitzone setzen
      saveflag = 1;                                                   // genderte Einstellungen im EEPROM speichern
    }
    cli();                                                            // Interrupts sperren
    hour = h;                                                         // Stunde setzen
    minute = m;                                                       // Minute setzen
    second = s;                                                       // Sekunde setzen
    intcnt = hs >> 1;                                                 // Hundertstel halbieren (in Bereich 0-49 wandeln) und setzen
    syncstat = 4;                                                     // Uhr ist jetzt vom Web-Modul synchronisiert
    synctout = c_sync;                                                // Synchronisierungs-Timeout neu starten
    sei();                                                            // Interrupts wieder freigeben
    lasthour = hour;                                                  // Stunde fr automatischen Farbwechsel speichern
    dcftout = 255;                                                    // DCF-Synchronisierungs-Timeout-Zhler stoppen
    matrix_on();                                                      // LED-Matrix wieder einschalten
  }
}

void msg_wm(void)                                                     // Textnachricht vom Web-Modul in den Pixelpuffer schreiben
{
  uint8_t  i = 3;                                                     // Zeigervariable auf Datenbereich der Empfangsdaten setzen
  uint16_t j = 0;                                                     // Zeigervariable auf Anfang des Pixelpuffers setzen
  char     c, c1;                                                     // Zeichen-Zwischenspeicher
  uint16_t offset;                                                    // Offset in der Zeichensatz-Tabelle
  uint8_t  cwidth;                                                    // Pixelbreite des Zeichens
  uint8_t  x;                                                         // Zhler fr die Zeichenausgabe, X-Position
  uint8_t  y;                                                         // Zhler fr die Zeichenausgabe, Y-Position
  uint8_t  b;                                                         // Zwischenspeicher fr Bytes aus der Zeichensatz-Tabelle
  uint8_t  s;                                                         // Byte mit in Spalten angeordneten Bits

  while (i < rxpos) {                                                 // solange noch Zeichen im Puffer sind
    c = rxdata[i ++];                                                 // Zeichen aus Empfangspuffer holen und Zeiger erhhen
    if ((c > 127) && (i < rxpos)) {                                   // wenn UTF-8-Zeichen und noch ein weiteres Zeichen im Puffer
      if (c < 0xE0) c = utf8_conv(c, rxdata[i ++]);                   // wenn 2-Byte-Zeichen -> Zeichen konvertieren (2 Byte)
      else {                                                          // sonst
        c1 = rxdata[i ++];                                            // zweites Zeichen holen
        if (i < rxpos) {                                              // wenn noch weitere Zeichen im Puffer
          c = utf8_conv3(c, c1, rxdata[i ++]);                        // Zeichen konvertieren (3 Byte)
        }
      }
    }
    if (c >= 32) {                                                    // wenn zulssiger Zeichencode
      offset = (c - 32) * 12;                                         // Tabellen-Offset berechnen
      cwidth = pgm_read_byte(&charset_normal_array[offset]);          // Zeichenbreite holen
      if (cwidth > 8) cwidth = 8;                                     // Zeichenbreite auf 8 begrenzen
      for (x = cwidth; x > 0; x--) {                                  // X-Pixelposition von Zeichenbreite bis 1 bearbeiten
        s = 0;                                                        // Spaltenbyte 1 lschen
        for (y = 0; y < 3; y ++) {                                    // 3 Zeilenbytes bearbeiten (Zeile 0-2)
          s <<= 1;                                                    // Spaltenbits nach links verschieben
          b = pgm_read_byte(&charset_normal_array[offset + y + 1]);   // Zeilenbyte holen
          s |= (b >> (x - 1)) & 1;                                    // bentigtes Bit an Position 0 schieben, filtern und zum Spaltenbyte hinzufgen
        }
        if (j < sizeof(msgpbuf)) msgpbuf[j ++] = s;                   // wenn noch Platz im Pixelpuffer -> Spaltenbyte 1 in Puffer schreiben
        s = 0;                                                        // Spaltenbyte 2 lschen
        for (y = 3; y < 11; y ++) {                                   // 8 Zeilenbytes bearbeiten (Zeile 3-10)
          s <<= 1;                                                    // Spaltenbits nach links verschieben
          b = pgm_read_byte(&charset_normal_array[offset + y + 1]);   // Zeilenbyte holen
          s |= (b >> (x - 1)) & 1;                                    // bentigtes Bit an Position 0 schieben, filtern und zum Spaltenbyte hinzufgen
        }
        if (j < sizeof(msgpbuf)) msgpbuf[j ++] = s;                   // wenn noch Platz im Pixelpuffer -> Spaltenbyte 2 in Puffer schreiben
      }
    }
    if (j < sizeof(msgpbuf)) msgpbuf[j ++] = 0;                       // wenn noch Platz im Pixelpuffer -> leeres Spaltenbyte 1 in Puffer schreiben (Zwischenraum)
    if (j < sizeof(msgpbuf)) msgpbuf[j ++] = 0;                       // wenn noch Platz im Pixelpuffer -> leeres Spaltenbyte 2 in Puffer schreiben (Zwischenraum)
  }
  msgpbend = j;                                                       // Nachrichtenende festlegen
  msgpbpos = -128;                                                    // Ausgabeposition auf 64 Pixelspalten vor den Anfang setzen
  if (msgpbend > 0) {                                                 // wenn eine Nachricht im Puffer
    if (msg_flag == 0) msg_flag = 1;                                  // wenn Nachrichten-Flag noch nicht gesetzt -> Nachrichten-Flag setzen
    if (syncstat && almsnd && sounflag && !mute) sndflags[4] = 1;     // wenn Uhr synchronisiert, akustische Alarme eingeschaltet, Soundzeit und keine Stummschaltung -> Sound-Flag 4 setzen
  }
}

#endif /* FUNCTIONS_H_ */
