/*
 * RGB-Matrix-Uhr 64x32, Scott-Falk Hhn, "interrupt.h"
 * Interrupt-Routinen
 */

#ifndef INTERRUPT_H_
#define INTERRUPT_H_

ISR(TIMER0_COMPA_vect, ISR_NOBLOCK)                                   // Timer0: Output Compare A Match Interrupt, steuert Taster und Timeout-Zhler, Routine unterbrechbar, Takt: 625Hz / 1,6ms
{
  msgscrol ++;                                                        // Nachrichten-Roll-Timeout-Zhler erhhen
  if (msgscrol >= 90 - msgscr * 4) {                                  // wenn berechneter Endwert erreicht (Bereich von 0-20 bei msgscr ergibt Zeiten von 144ms bis 16ms in Schritten von 6,4ms)
    msgscrol = 0;                                                     // Zhler zurcksetzen
    if (msgpbpos < msgpbend) msgpbpos += 2;                           // wenn Nachrichtenende noch nicht erreicht -> Zeiger auf nchste Pixelspalte (nchstes Bytepaar)
    else msgpbpos = -128;                                             // sonst auf 64 Pixelspalten vor den Anfang setzen
    msgstest--;                                                       // Testzhler vermindern (fr Einstellungs-Men von msgscr)
  }

  tim0cnt1 ++;                                                        // Interruptzhler 1 erhhen
  if (tim0cnt1 >= 25) {                                               // Endwert (24) berschritten?
    tim0cnt1 = 0;                                                     // ja -> zurcksetzen (Takt: 25Hz / 40ms)

    uint8_t keytemp;                                                  // Zwischenspeicher fr aktuellen Tasterstatus
    keytemp = ~PIND & _BV(PIND7);                                     // Tasterstatus 1 lesen, invertieren und filtern
    if (keytemp == key1old) {                                         // wenn aktueller = vorheriger Status
      if (keytemp) {                                                  // wenn Taster gedrckt
        key1stat = 1;                                                 // aktuellen Status auf "Gedrckt" setzen
        if (key1long < c_keyl1) key1long ++;                          // wenn Endwert fr langen Tastendruck noch nicht erreicht -> Zhler erhhen
      }
      else {                                                          // wenn Taster losgelassen
        key1stat = 0;                                                 // aktuellen Status auf "Losgelassen" setzen
        key1quit = 0;                                                 // Quittierungs-Status lschen
        key1long = 0;                                                 // Zhler fr langen Tastendruck lschen
      }
    }
    key1old = keytemp;                                                // Status fr nchste Taster-Abfrage speichern

    keytemp = ~PIND & _BV(PIND6);                                     // Tasterstatus 2 lesen, invertieren und filtern
    if (keytemp == key2old) {                                         // wenn aktueller = vorheriger Status
      if (keytemp) {                                                  // wenn Taster gedrckt
        key2stat = 1;                                                 // aktuellen Status auf "Gedrckt" setzen
        if (key2long < c_keyl1) key2long ++;                          // wenn Endwert fr langen Tastendruck noch nicht erreicht -> Zhler erhhen
      }
      else {                                                          // wenn Taster losgelassen
        key2stat = 0;                                                 // aktuellen Status auf "Losgelassen" setzen
        key2quit = 0;                                                 // Quittierungs-Status lschen
        key2long = 0;                                                 // Zhler fr langen Tastendruck lschen
        key2lcnt = 0;                                                 // Zykluszhler fr langen Tastendruck lschen
      }
    }
    key2old = keytemp;                                                // Status fr nchste Taster-Abfrage speichern

    keytemp = ~PIND & _BV(PIND5);                                     // Tasterstatus 3 lesen, invertieren und filtern
    if (keytemp == key3old) {                                         // wenn aktueller = vorheriger Status
      if (keytemp) key3stat = 1;                                      // wenn Taster gedrckt -> aktuellen Status auf "Gedrckt" setzen
      else {                                                          // wenn Taster losgelassen
        key3stat = 0;                                                 // aktuellen Status auf "Losgelassen" setzen
        key3quit = 0;                                                 // Quittierungs-Status lschen
      }
    }
    key3old = keytemp;                                                // Status fr nchste Taster-Abfrage speichern

    if (quitcnt < 32) quitcnt ++;                                     // wenn Quittierungszhler aktiv -> erhhen

    bright[bripos ++] = ADCH;                                         // neuen Helligkeitswert in Puffer schreiben, Position erhhen
    if (bripos > 31) bripos = 0;                                      // wenn letzte Position berschritten -> zurcksetzen

    int8_t i;                                                         // Schleifenzhler
    uint8_t m = 255;                                                  // Merker fr Suche nach gesetzten Sound-Flags auf ungltigen Wert setzen
    if (sndtout == 255) {                                             // wenn Soundausgabe nicht aktiv
      for (i = 5; i >= 0; i--) {                                      // 6 Sound-Flags rckwrts durchsuchen, damit Sound mit hchster Prioritt erkannt wird
        if (sndflags[i] > 0) m = i;                                   // wenn Flag gesetzt -> in Merker speichern
      }
      if (m < 6) {                                                    // wenn ein Sound-Flag gesetzt
        actsound = m;                                                 // auslsenden Sound speichern
        sndtout = 0;                                                  // Timeoutzhler starten
        switch (actsound) {                                           // Soundausgang abhngig vom aktiven Sound einschalten
          case 0: PORTB |= _BV(PINB1); break;                         // falls Sound 1 (Alarmsignal 1) -> PortB1 auf High
          case 1: PORTB |= _BV(PINB2); break;                         // falls Sound 2 (Alarmsignal 2) -> PortB2 auf High
          case 2: PORTB |= _BV(PINB3); break;                         // falls Sound 3 (Alarmsignal 3) -> PortB3 auf High
          case 3: PORTB |= _BV(PINB4); break;                         // falls Sound 4 (Alarmsignal 4) -> PortB4 auf High
          case 4: PORTB |= _BV(PINB5); break;                         // falls Sound 5 (Alarmsignal 5) -> PortB5 auf High
          case 5: PORTB |= _BV(PINB6); break;                         // falls Sound 6 (Stundensignal) -> PortB6 auf High
        }
      }
    }
    else {                                                            // wenn Soundausgabe aktiv
      sndtout ++;                                                     // Timeout-Zhler erhhen
      if (sndtout == c_sndto1) {                                      // wenn Zeit fr Impulsende erreicht
        PORTB &= ~(_BV(PINB6) | _BV(PINB5) | _BV(PINB4) |
          _BV(PINB3) | _BV(PINB2) | _BV(PINB1));                      // alle Soundausgnge (PortB1-PortB6) auf Low setzen
      }
      if (sndtout == c_sndto2) {                                      // wenn Wartezeitende erreicht
        sndtout = 255;                                                // Timeout-Zhler stoppen
        sndflags[actsound] = 0;                                       // auslsendes Sound-Flag lschen
      }
    }

    tim0cnt2 ++;                                                      // Interruptzhler 2 erhhen
    if (tim0cnt2 >= 25) {                                             // Endwert (24) berschritten?
      tim0cnt2 = 0;                                                   // ja -> zurcksetzen (Takt: 1Hz / 1s)

      if (sndena < 255) {                                             // wenn Sound-Timeout-Zhler aktiv
        if (sndena > 0) sndena--;                                     // wenn Sound-Timeout-Zhler noch nicht abgelaufen -> Timout-Zhler vermindern
        if (sndena == 0) {                                            // wenn Sound-Timeout-Zhler abgelaufen
          sndena = 255;                                               // Zhler stoppen
          PORTD &= ~_BV(PIND1);                                       // Soundmodul einschalten
        }
      }

      tim0sec ++;                                                     // Sekundenzhler erhhen
      if (tim0sec > 59) {                                             // Endwert (59) berschritten?
        tim0sec = 0;                                                  // ja -> zurcksetzen (Takt: 1min)
        if (dcftout < 255) {                                          // wenn DCF-Synchronisierungs-Timeout-Zhler aktiv
          if (dcftout > 0) dcftout--;                                 // wenn DCF-Synchronisierungs-Timeout-Zhler noch nicht abgelaufen -> Timeout-Zhler vermindern
        }
        tim0min ++;                                                   // Minutenzhler erhhen
        if (tim0min > 59) {                                           // Endwert (59) berschritten?
          tim0min = 0;                                                // ja -> zurcksetzen (Takt: 1h)
          if (synctout < 255) {                                       // wenn Synchronisierungs-Timeout-Zhler aktiv
            if (synctout > 0) synctout--;                             // wenn Synchronisierungs-Timeout-Zhler noch nicht abgelaufen -> Timeout-Zhler vermindern
          }
          if (mutetout < 255) {                                       // wenn Stummschaltungs-Timeout-Zhler aktiv
            if (mutetout > 0) mutetout--;                             // wenn Stummschaltungs-Timeout-Zhler noch nicht abgelaufen -> Timeout-Zhler vermindern
          }
        }
      }
    }
  }
}

ISR(TIMER1_COMPA_vect, ISR_BLOCK)                                     // Timer1: Output Compare A Match Interrupt, Ausgabe einer Pixelzeile auf die LED-Matrix, Routine nicht unterbrechbar,
{                                                                     // Matrix-Scan-Modus 1/16 Takt: 6400Hz / 156,25s; Matrix-Scan-Modus 1/8 Takt: 5312Hz / 188,25s
  matrix_out((uint16_t *) matrix[matnum], mulcnt, bripwm, matscn);    // eine Pixel-Zeile der aktiven LED-Matrix ausgeben (Assembler-Routine)
  mulcnt ++;                                                          // Multiplex-Zhler erhhen
}

ISR(TIMER2_OVF_vect, ISR_BLOCK)                                       // Timer2: Overflow Interrupt, schaltet die Anzeige nach Ausgabe einer Pixelzeile ab (fr PWM-Steuerung), Routine nicht
{                                                                     // unterbrechbar
  PORTA |= _BV(PINA5);                                                // OE auf High (LED-Ausgnge deaktivieren)
  TIMSK2 = 0;                                                         // Timer2: Overflow Interrupt wieder deaktivieren
}

ISR(TIMER3_COMPA_vect, ISR_NOBLOCK)                                   // Timer3: Output Compare A Match Interrupt, DCF77-Datenempfang und Zeitzhlung, Routine unterbrechbar, Takt: 50Hz / 20ms
{
  uint8_t bitval;                                                     // Wert eines empfangenen DCF77-Bits (0 oder 1)
  uint8_t dcfstat;                                                    // Zwischenspeicher fr DCF77-Status
  uint8_t dcftemp;                                                    // Zwischenspeicher fr die Bearbeitung der DCF77-Tabelle
  uint8_t mflag;                                                      // Merker fr Monatswechsel und Zwischenspeicher fr Datenfeldsuche
  int8_t i;                                                           // Schleifenzhler

  dcfcnt ++;                                                          // DCF77-Zhler erhhen
  dcfstat = (PINA & _BV(PINA6));                                      // aktuellen DCF77-Status lesen
  if (dcfinv) dcfstat = !dcfstat;                                     // falls Invertierung aktiviert -> invertieren

  switch (tstled) {                                                   // Test-LED-Status bearbeiten
    case 0:                                                           // falls Test-LED aus
      PORTD |= _BV(PIND4);                                            // High an PortD4 ausgeben (LED ausschalten)
      break;
    case 1:                                                           // falls Test-LED auf DCF gesetzt
      if (dcfstat) PORTD |= _BV(PIND4);                               // wenn DCF77-Status = 1 -> High an PortD4 ausgeben (LED ausschalten)
      else PORTD &= ~_BV(PIND4);                                      // sonst Low an PortD4 ausgeben (LED einschalten)
      break;
    case 2:                                                           // falls Test-LED auf RS-232 gesetzt
      if (tstlcnt < 10) {                                             // wenn Test-LED-Zhler aktiv
        tstlcnt ++;                                                   // Test-LED-Zhler erhhen
        PORTD &= ~_BV(PIND4);                                         // Low an PortD4 ausgeben (LED einschalten)
      }
      else PORTD |= _BV(PIND4);                                       // sonst High an PortD4 ausgeben (LED ausschalten)
      break;
  }

  if (!dcfstat1) {                                                    // wenn vorheriges DCF77-Signal Low
    if (dcfstat) {                                                    // wenn aktuelles DCF77-Signal High (= Impuls-Start)
      if ((dcfcnt > 2) && (dcfcnt < 13)) {                            // wenn Impulslnge 60-260ms -> gltiger Impuls
        dcfimp ++;                                                    // Impulsanzahl erhhen
        if (dcfcnt > 7) bitval = 1;                                   // wenn Impulslnge > 140ms -> 1-Bit
        else bitval = 0;                                              // sonst -> 0-Bit
        for (i = 41; i >= 0; i--) {                                   // 42 Pufferbytes bearbeiten (von hinten beginnend)
          dcftemp = dcftab[i];                                        // Pufferbyte holen und merken
          dcftab[i] = bitval;                                         // neues Pufferbyte setzen
          bitval = dcftemp;                                           // bertrag fr nchstes Pufferbyte
        }
      }
    }
  }
  else {                                                              // wenn vorheriges DCF77-Signal High
    if (!dcfstat) {                                                   // wenn aktuelles DCF77-Signal Low (= Impuls-Ende)
      if (dcfcnt > 75) dcfstat2 = 1;                                  // wenn Impulslnge > 1,5s -> neuer Minutenbeginn
      if (dcfcnt > 45) dcfcnt = 0;                                    // wenn Imuplslnge > 900ms -> Zhler auf 0 setzen
    }
  }
  dcfstat1 = dcfstat;                                                 // neuen DCF77-Status speichern

  intcnt ++;                                                          // Interruptzhler erhhen
  if (intcnt >= 50) {                                                 // Endwert (49) berschritten?
    intcnt = 0;                                                       // ja -> zurcksetzen (Takt: 1Hz, 1s)
    if (disptout < 255) {                                             // wenn Anzeige-Timeout-Zhler aktiv
      if (disptout > 0) disptout--;                                   // wenn Anzeige-Timeout-Zhler noch nicht abgelaufen -> Timeout-Zhler vermindern
    }
    if (nacttout < 255) {                                             // wenn Nachtaktivierungs-Zhler aktiv
      if (nacttout > 0) nacttout --;                                  // wenn Nachtaktivierungs-Zhler noch nicht abgelaufen -> Timeout-Zhler vermindern
    }

    second ++;                                                        // Sekunden erhhen
    if ((second % chtime) == (chtime - 2)) {                          // wenn Wechselzeit erreicht (bei 2 und Sekunde 0, 2, 4, ...; bei 3 und Sekunde 1, 4, 7, ...; bei 4 und Sekunde 2, 6, 10, ...)
      daymcoun ++;                                                    // Zhler fr Taganzeige erhhen
      if (daymcoun > 1) daymcoun = 0;                                 // wenn Endwert berschritten -> zurcksetzen

      for (i = 0; i < 31; i ++) {                                     // 31 Datenfeld-Timeout-Zhler bearbeiten
        if (sentout[i] < 255) {                                       // wenn Zhler aktiv
          if (sentout[i] > 0) sentout[i]--;                           // wenn Zhler noch nicht abgelaufen -> Zhler vermindern
          if (sentout[i] == 0) {                                      // wenn Zhler ablaufen
            sentout[i] = 255;                                         // Zhler stoppen
            senarray[i][0] = 0;                                       // Datenfeld lschen
          }
        }
      }

      datacoun ++;                                                    // Datenfeldzhler erhhen
      if (datacoun > 33) datacoun = 0;                                // wenn Datenfeldzhler auerhalb des zulssigen Bereiches
      if ((datacoun == 0) && !shdate) datacoun ++;                    // wenn Zhler auf Datum und dieses nicht angezeigt werden soll -> Zhler auf Wochentag
      if ((datacoun == 1) && !shwday) datacoun ++;                    // wenn Zhler auf Wochentag und dieser nicht angezeigt werden soll -> Zhler auf Sensorfeld
      mflag = datacoun;                                               // aktuellen Zhlerstand zwischenspeichern
      if (datacoun > 1) {                                             // wenn Zhler > 1 (Sensorwert oder Geburtstag)
        for (i = 0; i < 2; i ++) {                                    // nachfolgende Sensorfeldsuche zweimal durchfhren
          if (senfil) {                                               // wenn Sensorfilter eingeschaltet
            while (datacoun < 10) {                                   // wiederholen solange Zhler im zulssigen Bereich
              if ((sennum[datacoun - 2] > 0) && (senarray[sennum[datacoun - 2] - 1][0] > 0)) break; // wenn aktueller Eintrag eine Sensornummer enthlt und Datenfeld belegt ist -> Abbruch
              datacoun ++;                                            // Datenfeldzhler erhhen
            }
            if (datacoun > 9) datacoun = 0;                           // wenn alle Datenfelder ausgegeben -> Zhler zurcksetzen
          }
          else {                                                      // wenn Sensorfilter ausgeschaltet
            while (datacoun < 33) {                                   // wiederholen solange Zhler im zulssigen Bereich
              if (senarray[datacoun - 2][0] > 0) break;               // wenn belegtes Datenfeld gefunden -> Abbruch
              datacoun ++;                                            // Datenfeldzhler erhhen
            }
            if (datacoun > 32) datacoun = 0;                          // wenn alle Datenfelder ausgegeben -> Zhler zurcksetzen
          }
          if (i == 0) {                                               // wenn erster Such-Durchlauf
            if (!shdate && !shwday && (datacoun == 0) && (mflag > 2)) { // wenn Datum und Wochentag nicht angezeigt werden sollen, Datenfeldzhler zurckgesetzt und vorher ein Datenfeld angezeigt wurde
              datacoun = 2;                                           // Zhler fr erneute Suche auf erstes Datenfeld setzen
            }
            if (shdate || shwday) i ++;                               // wenn Datum oder Wochentag angezeigt werden sollen -> zweiten Such-Durchlauf berspringen
          }
        }
      }
      if ((datacoun == 0) && (mflag > 0)) {                           // wenn Zhler jetzt auf Datum und vorher ein anderes Feld angezeigt wurde
        if (!shdate) {                                                // wenn Datum nicht angezeigt werden soll
          datacoun ++;                                                // Zhler auf Wochentag
          if (!shwday) datacoun = 33;                                 // wenn Wochentag nicht angezeigt werden soll -> "keine Daten" anzeigen
        }
      }

      for (i = 0; i < 8; i ++) {                                      // 8 Alarmfeld-Timeout-Zhler bearbeiten
        if (almtout[i] < 255) {                                       // wenn Zhler aktiv
          if (almtout[i] > 0) almtout[i]--;                           // wenn Zhler noch nicht abgelaufen -> Zhler vermindern
          if (almtout[i] == 0) {                                      // wenn Zhler abgelaufen
            almtout[i] = 255;                                         // Zhler stoppen
            if (almhold[i] == 0) almarray[i][0] = 0;                  // wenn kein Halte-Flag gesetzt -> Alarm lschen
          }
        }
      }
      almcoun ++;                                                     // Alarmfeldzhler erhhen
      i = almcoun;                                                    // Zhler zwischenspeichern
      while (almcoun < 8) {                                           // wiederholen solange Zhler im zulssigen Bereich
        if (almarray[almcoun][0] > 0) break;                          // wenn belegtes Alarmfeld gefunden -> Abbruch
        almcoun ++;                                                   // sonst Alarmfeldzhler erhhen
      }
      if (almcoun > 7) {                                              // wenn alle Alarmfelder ausgegeben
        almcoun = 0;                                                  // Zhler zurcksetzen und
        while (almcoun < i) {                                         // weiter bis zum letzten Zhlerstand suchen
          if (almarray[almcoun][0] > 0) break;                        // wenn belegtes Alarmfeld gefunden -> Abbruch
          almcoun ++;                                                 // sonst Alarmfeldzhler erhhen
        }
      }
      alm_flag = 0;                                                   // Alarm-Flag zunchst ausschalten
      for (i = 0; i < 8; i ++) {                                      // Alarmfelder durchsuchen
        if (almarray[i][0] > 0) alm_flag = 1;                         // wenn Alarm-Eintrag gefunden -> Alarm-Flag setzen
      }
    }

    mflag = 0;                                                        // Merker fr Monatswechsel lschen
    if (second > 59) {                                                // wenn berlauf
      second = 0;                                                     // Sekunden zurcksetzen (Takt: 1min)
      minute ++;                                                      // Minuten erhhen
      if (minute > 59) {                                              // wenn berlauf
        minute = 0;                                                   // Minuten zurcksetzen (Takt: 1h)
        hour ++;                                                      // Stunden erhhen
        if (autost) {                                                 // wenn Sommerzeit-Automatik eingeschaltet
          if ((day > 24) && (wday == 7)) {                            // wenn Wechsel-Zeitraum und Sonntag
            if ((month == 3) && (hour == 2) && (timezone == 0)) {     // wenn Mrz, 2:00 Uhr und noch Normalzeit
              hour ++;                                                // eine Stunde vorstellen
              timezone = 1;                                           // Zeitzone auf MESZ setzen
              saveflag = 1;                                           // genderte Zeitzone speichern
            }
            if ((month == 10) && (hour == 3) && (timezone == 1)) {    // wenn Oktober, 3:00 Uhr und noch Sommerzeit
              hour --;                                                // eine Stunde zurckstellen
              timezone = 0;                                           // Zeitzone auf MEZ setzen
              saveflag = 1;                                           // genderte Zeitzone speichern
            }
          }
        }
        if (hour > 23) {                                              // wenn berlauf
          hour = 0;                                                   // Stunden zurcksetzen (Takt: 1d)
          wday ++;                                                    // Wochentag erhhen
          if (wday > 7) wday = 1;                                     // wenn Wochentagberlauf -> Wochentag zurcksetzen
          day ++;                                                     // Tag erhhen
          if ((year % 4) == 0) {                                      // wenn Schaltjahr
            if (day > pgm_read_byte(&daytab2[month - 1])) mflag = 1;  // wenn Tag > Eintrag in Schaltjahr-Tabelle -> Monatswechsel
          }
          else {                                                      // wenn kein Schaltjahr
            if (day > pgm_read_byte(&daytab1[month - 1])) mflag = 1;  // wenn Tag > Eintrag in Normaljahr-Tabelle -> Monatswechsel
          }
          if (mflag) {                                                // wenn Monatswechsel erforderlich
            day = 1;                                                  // Tag zurcksetzen (Takt: 1 Monat)
            month ++;                                                 // Monat erhhen
            if (month > 12) {                                         // wenn berlauf
              month = 1;                                              // Monat zurcksetzen (Takt: 1 Jahr)
              year ++;                                                // Jahr erhhen
            }
          }
        }
      }
    }
  }
}

ISR(USART1_RX_vect, ISR_BLOCK)                                        // USART1 RX Complete, empfangenes Zeichen wird im RX-Puffer abgelegt, Routine nicht unterbrechbar
{
  rxbuff[rxwrite ++] = UDR1;                                          // empfangenes Zeichen auf Zeigerposition in den RX-Puffer schreiben und Schreib-Zeiger erhhen
}

#endif /* INTERRUPT_H_ */
